import pytest
from django.http import HttpResponse
from pytest_djangoapp import configure_djangoapp_plugin
from pytest_djangoapp.fixtures.request import DjagoappClient

from siteajax.utils import Source


pytest_plugins = configure_djangoapp_plugin(
    extend_MIDDLEWARE=[
        'siteajax.middleware.ajax_handler',
    ],
)


class HtmxMock(DjagoappClient):

    @classmethod
    def get_extra(
        cls,
        source: Source = None,
        target: str = '',
        user_input: str = '',
        url: str = 'http://testhost/dummy/',
    ):
        extra = {
            'HTTP_HX-Request': 'true',
            'HTTP_HX-Current-Url': url,
        }

        if source:
            extra['HTTP_HX-Trigger'] = source.id
            extra['HTTP_HX-Trigger-Name'] = source.name

        if target:
            extra['HTTP_HX-Target'] = target

        if user_input:
            extra['HTTP_HX-Prompt'] = user_input

        return extra

    @classmethod
    def make_headers_a(cls, response: HttpResponse):

        headers = getattr(response, 'headers', getattr(response, '_headers', None))
        # _headers - in pre Django 3.2

        headers_a = {}

        for key, val in headers.items():
            if key.lower().startswith('hx'):
                if isinstance(val, tuple):  # pre 3.2
                    key, val = val[0], val[1]
                headers_a[key] = val

        response.headers_a = headers_a

    def get(
        self,
        path: str,
        data=None,
        source: Source = None,
        target: str = '',
        user_input: str = '',
        **kwargs
    ):
        extra = self.get_extra(
            source=source,
            target=target,
            user_input=user_input,
        )
        response = super().get(path, data=data, follow=False, secure=False, **extra, **kwargs)
        self.make_headers_a(response)

        return response


@pytest.fixture
def htmx():

    def request_client_(user=None, raise_exceptions=True, json=False, **kwargs) -> HtmxMock:
        return HtmxMock(
            user=user,
            raise_exceptions=raise_exceptions,
            json=json,
            **kwargs)

    return request_client_



from typing import Callable, Dict

from django.http import HttpResponse

from .utils import Ajax


def ajax_dispatch(views_map: Dict[str, Callable]) -> Callable:
    """Decorator. Allows ajax request dispatch based 
    on source html element identifiers.

    Useful in cases when various ajax calls have a single entry point view.

    .. code-block::
        // JS. Define an element with `id`.
        <button id="replaceit" hx-get="/">Push me</button>

        // Python. Define a handling view and an entry point view decorated.

        def do_replace(request):
            return HttpResponse('Thank you')

        @ajax_dispatch({
            'replaceit': do_replace,  # Map element id to a handler
            'prefixed-some-*': do_replace,  # Map several id with the same prefix
        })
        def index(request):
            return HttpResponse('not ajax')
     
    :param views_map: Map html elements IDs to handling functions.
        To match several IDs use start (*)

    """
    wildcards = {}
    ids = {}

    for key, handler in views_map.items():
        prefix, sep, _ = key.partition('*')

        if sep == '*':
            wildcards[prefix] = handler

        else:
            ids[key] = handler

    def ajax_view_(func: Callable) -> Callable:

        def view_wrapper(*fargs, **fkwargs) -> HttpResponse:

            request = fargs[0]

            if not hasattr(request, 'POST'):
                # Possibly a class-based view where 0-attr is `self`.
                request = fargs[1]

            if hasattr(request, 'ajax'):
                # Attribute is already set by middleware.
                ajax: Ajax = request.ajax

            else:
                # Initialize on fly.
                ajax = Ajax(request)
                request.ajax = ajax

            if ajax:
                source_id = ajax.source.id
                handling_view = ids.get(source_id)

                if not handling_view:
                    # Long way to process wildcards.
                    for id_prefix, handling in wildcards.items():
                        if source_id.startswith(id_prefix):
                            handling_view = handling
                            break

                if handling_view:
                    response = handling_view(*fargs, **fkwargs)
                    # Try to unwrap an AjaxResponse if any.
                    return getattr(response, 'wrapped_response', response)

            return func(*fargs, **fkwargs)

        return view_wrapper

    return ajax_view_





def test_template(request_client):
    client = request_client()

    response = client.get('/sample_view_1/').content.decode()
    assert "'htmx:configRequest'" in response
    assert '<script src="https://unpkg.com/' in response



from siteajax.utils import Ajax


def test_event_header(request_get):
    request = request_get(**{
        'HTTP_Triggering-Event-URI-AutoEncoded': 'true',
        'HTTP_Triggering-Event': '%7B%0A%20%22isTrusted%22%3A%20true%2C%0A%20%22BUBBLING_PHASE%22%3A%203%0A%7D'
    })

    ajax = Ajax(request)
    event_data = ajax.event
    assert event_data == {'isTrusted': True, 'BUBBLING_PHASE': 3}

    # no data
    assert Ajax(request_get()).event == {}



from django.http import HttpResponse, HttpRequest
from django.shortcuts import redirect, render
from django.views import View
from siteajax.toolbox import Ajax, AjaxResponse, ajax_dispatch


def sample_view_1(request: HttpRequest) -> HttpResponse:

    ajax: Ajax = request.ajax

    if ajax:
        do_redirect = request.GET.get('doredir') == '1'
        do_events = request.GET.get('doevents') == '1'

        if do_redirect:
            response = redirect('/somwhere/')

        else:
            response = HttpResponse('ajaxresp')

        response = AjaxResponse(response)

        response.history_item = '/otherurl/'
        response.redirect = '/here/'
        response.refresh = True

        if do_events:
            response.trigger_event(name='fireThis', kwargs={'one': {'two': 3}})
            response.trigger_event(name='fireThat')
            response.trigger_event(name='fireMe1', kwargs={'me': True}, step='swap')
            response.trigger_event(name='fireMe2', step='settle')

        return response

    return render(request, 'index.html')


def get_dispatched(request: HttpRequest) -> HttpResponse:
    return HttpResponse('dispatched')


@ajax_dispatch({
    'dispatchme': get_dispatched
})
def sample_view_2(request: HttpRequest) -> HttpResponse:
    return HttpResponse('notajax')


class SampleView3(View):

    def get_dispatched_2(self, request: HttpRequest) -> HttpResponse:
        return HttpResponse('dispatched')

    @ajax_dispatch({
        'dispatchme': get_dispatched_2
    })
    def get(self, request: HttpRequest) -> HttpResponse:
        return HttpResponse('notajax')



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SiteajaxConfig(AppConfig):
    """Application configuration."""

    name = 'siteajax'
    verbose_name = _('Siteajax')



from typing import Callable

from django.http import HttpRequest, HttpResponse

from .toolbox import Ajax


def ajax_handler(get_response: Callable) -> Callable:
    """This handles a request - response to making
    it easy to handle Ajax.

    :param get_response:

    """
    def handle(request: HttpRequest) -> HttpResponse:

        request.ajax = Ajax(request)

        response = get_response(request)

        # Try to unwrap an AjaxResponse if any.
        return getattr(response, 'wrapped_response', response)

    return handle



from django.http import HttpRequest, HttpResponse

from siteajax.decorators import ajax_dispatch
from siteajax.utils import Source, AjaxResponse


def test_headers(htmx):
    client = htmx()

    response = client.get(
        '/sample_view_1/',
        source=Source(id='myid', name='myname'),
        target='trgt',
        user_input='here i am',
        data={'doevents': 1}
    )

    assert response.status_code == 200
    assert response.headers_a == {
        'HX-Push': '/otherurl/',
        'HX-Redirect': '/here/',
        'HX-Refresh': 'true',
        'HX-Trigger': '{"fireThis": {"one": {"two": 3}}, "fireThat": {}}',
        'HX-Trigger-After-Settle': '{"fireMe2": {}}',
        'HX-Trigger-After-Swap': '{"fireMe1": {"me": true}}',
    }

    response = client.get(
        '/sample_view_1/',
        data={'doredir': 1}
    )
    assert response.status_code == 302
    assert response.headers_a == {'HX-Push': '/otherurl/', 'HX-Redirect': '/here/', 'HX-Refresh': 'true'}


def test_dispatch(htmx):
    client = htmx()

    def do_test(url: str):
        response = client.get(url, source=Source(id='dispatchme', name=''))
        assert response.content.decode() == 'dispatched'

        # not dispatched
        response = client.get(url, source=Source(id='unknown', name=''))
        assert response.content.decode() == 'notajax'

    # function based view
    do_test('/sample_view_2/')

    # class based view
    do_test('/sample_view_3/')


def test_decor_ajax_autoinit(request_get):

    def handler(request: HttpRequest):
        response = AjaxResponse(HttpResponse('fine'), js_redirect=True)
        response.history_item = 'addthis/'
        return response

    @ajax_dispatch({'me': handler})
    def my_view(request: HttpRequest):
        return HttpResponse('nope')

    request = request_get(**{
        'HTTP_HX-Request': 'true',
        'HTTP_HX-Trigger': 'me',
    })

    response = my_view(request)
    assert response.content == b'fine'
    assert response['HX-Push'] == 'addthis/'


def test_decor_wildcards(request_get):

    def handler(request: HttpRequest):
        return HttpResponse(request.ajax.source.id)

    @ajax_dispatch({
        'my-not': lambda request: HttpResponse('miss'),
        'my-id-*': handler,
    })
    def my_view(request: HttpRequest):
        return HttpResponse('nope')

    request = request_get(**{
        'HTTP_HX-Request': 'true',
        'HTTP_HX-Trigger': 'my-id-1234567',
    })

    response = my_view(request)
    assert response.content == b'my-id-1234567'



from django.urls import path

from . import views

urlpatterns = [
    path('sample_view_1/', views.sample_view_1),
    path('sample_view_2/', views.sample_view_2),
    path('sample_view_3/', views.SampleView3.as_view()),
]



from json import dumps, loads
from typing import NamedTuple, Dict, Any
from urllib.parse import unquote

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpRequest, HttpResponse
from django.http.response import HttpResponseRedirectBase


class Source(NamedTuple):
    """Describes an element sourcing (triggering) a request."""

    id: str
    """The id of the triggered element if exists."""

    name: str
    """The name of the triggered element if exists."""


class Ajax:
    """Describes additional data sent with Ajax request.

    .. note:: The object is lazily initialized to allow faster
        middleware processing.

        Without initialization you won't be able to access it's attributes.
        For initialization it's enough to check it in boolean context, e.g.::

            bool(Ajax(request))

            # or

            if request.ajax:
                ...

    """
    __slots__ = ['is_used', 'url', 'source', 'target', 'user_input', 'restore_history', '_request']

    def __init__(self, request: HttpRequest):
        self._request = request

    def _boot(self):
        headers = self._request.headers

        self.is_used: bool = headers.get('Hx-Request', '') == 'true'
        """Indicates whether Ajax request is issued."""

        self.restore_history: bool = headers.get('HX-History-Restore-Request', '') == 'true'
        """Indicates the client side request to get the entire page 
        (as opposed to a page fragment request), when the client was 
        unable to restore a browser history state from the cache.
        
        """

        self.url: str = headers.get('Hx-Current-Url', '')
        """The current URL of the browser."""

        self.source: Source = Source(
            id=headers.get('HX-Trigger', ''),
            name=headers.get('HX-Trigger-Name', ''),
        )
        """Describes an element sourcing (triggering) a request."""

        self.target: str = headers.get('HX-Target', '')
        """The id of the target element if it exists."""

        self.user_input: str = headers.get('HX-Prompt', '')
        """The user input to a prompt (hx-prompt)."""

    def __bool__(self):
        self._boot()
        return self.is_used

    @property
    def event(self) -> dict:
        """Returns a dictionary describing a triggering event.

        Requires `event-header` extension:
            https://htmx.org/extensions/event-header/

        """
        headers = self._request.headers

        data = headers.get('Triggering-Event', '')
        if not data:
            return {}

        # encoded = headers.get('Triggering-Event-Uri-Autoencoded', '') == 'true'
        data = unquote(data)

        return loads(data)


class AjaxResponse(HttpResponse):
    """Represents a response object capable of driving a client side."""

    __slots__ = [
        '_wrapped', '_headers', '_triggers',
        'js_redirect',
        'history_item', 'redirect', 'refresh',
    ]

    _trigger_stages = {
        'receive': 'HX-Trigger',
        'settle': 'HX-Trigger-After-Settle',
        'swap': 'HX-Trigger-After-Swap',
    }

    def __init__(self, response: HttpResponse, *, js_redirect: bool = True, **kwargs):
        """

        :param response: Base response object.

        :param js_redirect: Whether to convert a redirect response object
            into an instruction for a client js library.

            * True - redirect is handled by a client side js library. Js library
                will get a result from this response.

            * False - redirect is handled by a browser. Js library will get
                the result from an URL browser has redirected it to.

        """
        super().__init__(**kwargs)

        self._wrapped = response
        self.js_redirect: bool = js_redirect

        self._triggers: Dict[str, Dict[str, Any]] = {
            'receive': {},
            'settle': {},
            'swap': {},
        }

        self.history_item: str = ''
        """Allows to push a new url into the browser history stack."""

        self.redirect: str = ''
        """Instructs a client-side redirect to a new location."""

        self.refresh: bool = False
        """Instructs a client side to make full refresh of the page."""

    @property
    def wrapped_response(self) -> HttpResponse:
        """Returns an base response modified to allow client driving."""

        response = self._wrapped
        headers = getattr(response, 'headers', None)

        if headers is None:  # pragma: nocover
            # pre Django 3.2
            headers = response

        val = self.history_item
        if val:
            headers['HX-Push'] = val

        val = self.redirect
        if val:
            headers['HX-Redirect'] = val

        if self.redirect:
            headers['HX-Refresh'] = 'true'

        if self.js_redirect and isinstance(response, HttpResponseRedirectBase):
            self.redirect = response.url
            del headers['location']  # Do not trigger browser redirect

        # Now encode event triggers data.
        trigger_stages = self._trigger_stages

        for stage, events in self._triggers.items():

            if not events:
                continue

            header_key = trigger_stages.get(stage)

            if header_key:
                headers[header_key] = dumps(events, cls=DjangoJSONEncoder)

        return response

    def trigger_event(self, *, name: str, kwargs: Dict[str, Any] = None, step: str = 'receive'):
        """Can be used to trigger client side actions on the target element within a response.

        .. code-block::
            // Python
            response.trigger_event(name='myEvent', kwargs={'one': {'two': 3}})

            // JS
            document.body.addEventListener('myEvent', function(event){
                console.log(event.detail.one.two);
            })

        :param name: Event name to trigger.

        :param kwargs: Keyword arguments to pass to an event.
            Those will be available from event.detail object.

        :param step: When to trigger this event.

            https://htmx.org/docs/#request-operations

            * receive - trigger events as soon as the response is received. Default.
            * settle - trigger events after the settling step.
            * swap - trigger events after the swap step.

        """
        self._triggers[step][name] = kwargs or {}
